/* FIXME: Add SOCK_STREAM support: #define USE_SOCK_DGRAM */
/* scm_righs+creds_recv.c

   Copyright (C) Svante Signell, 2015.

   Used in conjunction with scm_rights+creds_send.c to demonstrate
   passing of file descriptors and process credentials via a UNIX
   domain socket.

   This program receives file desciptors and credentials sent to a
   UNIX domain socket.

   File descriptors and process credentials are exchanged over
   datagram sockets.

   Modified for kFreeBSD and Hurd by Svante Signell April-Aug 2013.
*/

#define _GNU_SOURCE             /* To get SCM_CREDENTIALS definition from
                                   <sys/socket.h> */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <limits.h>
#include <fcntl.h>

#define SOCK_PATH "scm_rights+creds.socket"
#define BUF_SIZE 100
#define FD_MAX 3
#define USE_SOCK_DGRAM

int
main(int argc, char *argv[])
{
  struct msghdr msgh;
  struct iovec iov;
  int data;
  int i, j, k;
  int sfd;
  struct sockaddr_un addr;
  error_t err;
  int NRIGHTS = 1, NCREDS = 1;
  char line[BUF_SIZE];

  int *fdptr = NULL, *end = NULL;
  int fd[FD_MAX];

#if defined (__linux__)
  struct ucred ucred, *ucredp;
  int optval;
  socklen_t len;
#else
  struct cmsgcred *ucredp;
#endif
  int sz =  sizeof(*ucredp);
  struct cmsghdr *cmhp = NULL;
  ssize_t nr;

  printf ("Number of SCM_RIGHTS [<=3], SCM_CREDS [<=2]: [%d,%d] ", NCREDS, NRIGHTS);

  if (fgets(line, sizeof(line), stdin))
    {
      if (sscanf(line, "%d,%d", &NRIGHTS, &NCREDS) != 2)
	printf ("Input error: Using defaults: ");
    }
  printf ("\nNRIGHTS = %d, NCREDS = %d\n", NRIGHTS, NCREDS);

  /* Space large enough to hold NCREDS ucred structures + NRIGHTS FDs */
  int cspc1 = CMSG_SPACE (sizeof(int));
  int cspc2 = CMSG_SPACE (sz);
  int cspc = NRIGHTS * cspc1 + NCREDS * cspc2;
  if (cspc == 0)
    cspc = CMSG_SPACE(0);
  int clen1 = CMSG_LEN (sizeof(int));
  int clen2 = CMSG_LEN (sz);

  union {
    struct cmsghdr cmh;
    char control[cspc];
  } control_un;
  /* Set 'msgh' fields to describe 'control_un' */
  msgh.msg_control = control_un.control;
  msgh.msg_controllen = sizeof(control_un.control);
  printf ("scm_rights+creds_send.c: msgh.msg_controllen = %d (from control_un)\n", (int)msgh.msg_controllen);

  /* http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=514917 */
  /* FIXME: CMSG_NXTHDR(&msgh, cmhp) returns NULL if not zeroed out */
  memset (&control_un, 0, sizeof(control_un));

  /* Create socket bound to well-known address */
  if (remove(SOCK_PATH) == -1 && errno != ENOENT)
    {
      perror("remove");
      exit(EXIT_FAILURE);
    }

  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, SOCK_PATH, sizeof(addr.sun_path) - 1);
  sfd = socket (AF_UNIX, SOCK_DGRAM, 0);
  if (sfd == -1)
    {
      perror("socket");
      exit(EXIT_FAILURE);
    }

  err = bind (sfd, (struct sockaddr *) &addr, sizeof(struct sockaddr_un));
  if (err == -1)
    {
      perror("bind");
      exit(EXIT_FAILURE);
    }

  printf ("Receiving via datagram socket\n");

  /* We must set the SO_PASSCRED socket option in order to receive
     credentials */
#if defined(__linux__)
  if (NCREDS)
    {
      /* SO_PASSCRED not defined on GNU */
      optval = 1;
      if (setsockopt(sfd, SOL_SOCKET, SO_PASSCRED, &optval, sizeof(optval)) == -1)
	{
	  perror("setsockopt");
	  exit(EXIT_FAILURE);
	}
    }
#endif

  /* Set fields of 'msgh' to point to buffer used to receive (real)
     data read by recvmsg() */
  msgh.msg_iov = &iov;
  msgh.msg_iovlen = 1;
  iov.iov_base = &data;
  iov.iov_len = sizeof(int);

  msgh.msg_name = NULL;               /* We don't need address of peer */
  msgh.msg_namelen = 0;

  /* Receive real plus ancillary data */
  nr = recvmsg(sfd, &msgh, 0);
  if (nr == -1)
    {
      perror("recvmsg");
      exit(EXIT_FAILURE);
    }
  printf ("recvmsg() returned %ld\n", (long) nr);

  if (nr > 0)
    printf ("Received data = %d\n", data);

  printf("scm_rights+creds_recv.c: msgh.msg_controllen = %d\n",(int)msgh.msg_controllen);

  i = j = 0;
  for (cmhp = CMSG_FIRSTHDR(&msgh);
       cmhp != NULL;
       cmhp = CMSG_NXTHDR(&msgh, cmhp))
    {
      if (cmhp == NULL)
	{
	  perror("bad cmsg header cmhp == NULL");
	  exit(EXIT_FAILURE);
	}
      printf("scm_rights+creds_recv.c: cmhp[%d]->cmsg_len = %d\n", i+j, (int)cmhp->cmsg_len);
      if (cmhp->cmsg_level != SOL_SOCKET)
	{
	  perror("cmsg_level != SOL_SOCKET");
	  exit(EXIT_FAILURE);
	}

      /* Get the received credentials */
#ifdef __linux__
      if (cmhp->cmsg_type != SCM_CREDENTIALS && cmhp->cmsg_type != SCM_RIGHTS)
	{
	  printf("cmsg_type != SCM_CREDENTIALS && cmsg_type != SCM_RIGHTS\n");
	  exit(EXIT_FAILURE);
	}
#else
      if (cmhp->cmsg_type != SCM_CREDS && cmhp->cmsg_type != SCM_RIGHTS)
	{
	  printf("cmsg_type != SCM_CREDS && cmsg_type != SCM_RIGHTS\n");
	  exit(EXIT_FAILURE);
	}
#endif

      /* Get the received file descriptor (which is typically a
	 different file descriptor number than was used in the sending
	 process) */
      if (cmhp->cmsg_type == SCM_RIGHTS)
	{
	  printf ("\nscm_rights+creds_recv.c: Receiving SCM_RIGHTS data\n");
	  printf ("Received file descriptors: ");
	  //fd[i] = *(int *)CMSG_DATA(cmhp);
	  fdptr = (int *) CMSG_DATA(cmhp);
	  end = (int *)((char *)cmhp + cmhp->cmsg_len);
	  //for (k=0; &fdptr[k] < end; k++)
	  for (k=0; fdptr+k < end; k++)
	    printf("fd[%d] = %d ", i + k, fdptr[k]);
	  printf("\n");

	  /* Having obtained the file descriptor, read the file's
	     contents and print them on standard output */
	  for (k=0; fdptr+k < end; k++)
	    {
	      char buf[BUF_SIZE];
	      ssize_t numRead;

	      numRead = read(fdptr[k], buf, BUF_SIZE);
	      if (numRead == -1)
		{
		  perror("read");
		  exit(EXIT_FAILURE);
		}

	      if (numRead == 0)
		break;
	      printf("\nContents fd[%d]: \n", k);
	      write(STDOUT_FILENO, buf, numRead);
	      printf("\n");
	    }
	  i++;
	  continue;
	} /* SCM_RIGHTS */

#ifdef __linux__
	  if (cmhp->cmsg_type == SCM_CREDENTIALS)
	    {
	      printf ("\nscm_rights+creds_recv.c: Receiving SCM_CREDENTIALS data\n");
	      if (cmhp->cmsg_len != clen2)
		{
		  printf("bad ucred message length cmhp->cmsg_len = %d, "
			 "CMSG_LEN(sizeof(struct ucred)) = %d\n",
			 cmhp->cmsg_len, clen2);
		  exit(EXIT_FAILURE);
		}
	      ucredp = (struct ucred *) CMSG_DATA(cmhp);
	      printf("Received credentials[%d]: pid=%ld, uid=%ld, gid=%ld\n",
		     j,
		     (long) ucredp->pid, (long) ucredp->uid,
		     (long) ucredp->gid);
	      j++;
	      continue;
	    } /* SCM_CREDENTIALS */
#else /* GNU and kFreeBSD */
	  if (cmhp->cmsg_type == SCM_CREDS)
	    {
	      printf ("\nscm_rights+creds_recv.c: Receiving SCM_CREDS data\n");
	      if (cmhp->cmsg_len != clen2)
		{
		  printf("bad cmsg message length cmhp->cmsg_len = %d, "
			 "CMSG_LEN(sizeof(struct cmsgcred)) = %d\n",
			 cmhp->cmsg_len, clen2);
		  exit(EXIT_FAILURE);
		}
	      ucredp = (struct cmsgcred *) CMSG_DATA(cmhp);
	      printf("Received credentials[%d] pid=%ld, uid=%ld, euid=%ld, gid=%ld\n",
		     j,
		     (long) ucredp->cmcred_pid, (long) ucredp->cmcred_uid,
		     (long) ucredp->cmcred_euid, (long) ucredp->cmcred_gid);
	      for (k=0; k<ucredp->cmcred_ngroups; k++)
		printf("ucredp->cmcred_groups[%d]=%d\n", k, ucredp->cmcred_groups[k]);
	      j++;
	      continue;
	    } /* SCM_CREDS */
#endif /* __linux__ */
	} /* for (cmhp = ...) */

#if defined(__linux__) && !defined(USE_SOCK_DGRAM)
  if (NCREDS)
    {
      /* The Linux-specific, read-only SO_PEERCRED socket option
	 returns credential information about the peer, as described
	 in socket(7) */
      len = sizeof(struct ucred);
      if (getsockopt(sfd, SOL_SOCKET, SO_PEERCRED, &ucred, &len) == -1)
	perror("getsockopt");
      printf("WARNING: SO_PEERCRED does not work for datagram sockets\n");
      printf("WARNING: Does only work for stream sockets and datagram socket pairs created with socketpair\n");
      printf("Credentials from SO_PEERCRED: pid=%ld, euid=%ld, egid=%ld\n",
	     (long) ucred.pid, (long) ucred.uid, (long) ucred.gid);
    }
#endif

  exit(EXIT_SUCCESS);
}
