/* FIXME: Add SOCK_STREAM support: #define USE_SOCK_DGRAM */
/* scm_righs+creds_send.c

   Copyright (C) Svante Signell, 2015.

   Used in conjunction with scm_rights+creds_recv.c to demonstrate
   passing of file descriptors and process credentials via a UNIX
   domain socket.

   This program sends file descriptors and credentials to a UNIX
   domain socket.

   Credentials are exchanged over datagram sockets.

*/

#define _GNU_SOURCE             /* To get SCM_CREDENTIALS definition from
                                   <sys/socket.h> */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <limits.h>
#include <fcntl.h>

#define SOCK_PATH "scm_rights+creds.socket"
#define FD_MAX 3
#define BUF_SIZE 100
#define USE_SOCK_DGRAM

/*
For Linux it does not matter if '\0' or something else is
sent. Credentials passing is controlled by setsockopt option
SO_PASSCRED used in scm_rigths+creds_recv.c
 */

int
main(int argc, char *argv[])
{
  struct msghdr msgh;
  struct iovec iov;
  int sfd;
  struct sockaddr_un addr = { 0 };
  int data;
  int i, j, sum_controllen;
  int NRIGHTS = 1, NCREDS = 1;
  char line[BUF_SIZE];

  ssize_t ns;
  struct cmsghdr *cmhp = NULL;

  int *fdptr = NULL, fd[FD_MAX];
  char *filename[FD_MAX] = {"./test1.txt", "./test2.txt", "./test3.txt"};

  printf ("Number of SCM_RIGHTS [<=3], SCM_CREDS [<=2]: [%d,%d] ", NCREDS, NRIGHTS);
  if (fgets(line, sizeof(line), stdin))
    {
      if (sscanf(line, "%d,%d", &NRIGHTS, &NCREDS) != 2)
	printf ("Input error: Using defaults: ");
    }
  printf ("\nNRIGHTS = %d, NCREDS = %d\n", NRIGHTS, NCREDS);

#ifdef __linux__
  /* NOTE: ucred does not have euid defined in the struct */
  struct ucred *ucp;
#else
  struct cmsgcred *ucp;
#endif
  int sz =  sizeof(*ucp);

  /* Space large enough to hold NCREDS ucred structures + NRIGHTS FDs */
  int cspc1 = CMSG_SPACE (sizeof(int));
  int cspc2 = CMSG_SPACE (sz);
  int cspc = NRIGHTS * cspc1 + NCREDS * cspc2;
  if (cspc == 0)
    cspc = CMSG_SPACE(0);
  int clen1 = CMSG_LEN (sizeof(int));
  int clen2 = CMSG_LEN (sz);

  union {
  struct cmsghdr cmh;
  char control[cspc];
} control_un;

  /* Set 'msgh' fields to describe 'control_un' */
  msgh.msg_control = control_un.control;
  msgh.msg_controllen = sizeof(control_un.control);
  printf ("scm_rights+creds_send.c: msgh.msg_controllen = %d (from control_un)\n", (int)msgh.msg_controllen);

  /* http://bugs.debian.org/cgi-bin/bugreport.cgi?bug=514917 */
  /* FIXME: CMSG_NXTHDR(&msgh, cmhp) returns NULL if not zeroed out */
  memset (&control_un, 0, sizeof(control_un));

/* On Linux, we must transmit at least 1 byte of real data in
     order to send ancillary data */

  msgh.msg_iov = &iov;
  msgh.msg_iovlen = 1;

  /* Send default integer 12345 or taken from the command line */
  data = 12345;
  iov.iov_base = &data;
  iov.iov_len = sizeof(int);
  printf("Sent data = %d\n", data);

  /* Don't need to specify destination address, because we use
     connect() below */
  msgh.msg_name = NULL;
  msgh.msg_namelen = 0;

  /* Set message header to describe ancillary data that we want to
     send */
  i = j = 0;
  sum_controllen = 0;
  for (cmhp = CMSG_FIRSTHDR (&msgh);
       cmhp != NULL;
       cmhp = CMSG_NXTHDR (&msgh, cmhp))
    {
      if (cmhp == NULL)
	{
	  perror("cmhp == NULL");
	  exit(EXIT_FAILURE);
	}
      cmhp->cmsg_level = SOL_SOCKET;

      /* Set message header to describe ancillary data: NRIGHTS
	 FDs that we want to send */
      if ( i < NRIGHTS)
	{
	  printf ("\nscm_rights+creds_send.c: Sending SCM_RIGHTS data\n");
	  cmhp->cmsg_type = SCM_RIGHTS;
	  cmhp->cmsg_len = clen1;
	  printf ("scm_rights+creds_send.c: cmhp[%d]->cmsg_len = %d\n", i+j, clen1);
	  printf ("scm_rights+creds_send.c: CMSG_SPACE (sizeof(int))[%d] = %d\n", i+j, cspc1);
	  sum_controllen += cspc1;
	  fd[i] = open(filename[i], O_RDONLY);
	  if (fd[i] == -1)
	    {
	      perror("open fd[i]");
	      exit(EXIT_FAILURE);
	    }
	  printf ("scm_rights+creds_send.c: fd[%d]=%d\n", i, fd[i]);
	  *((int *) CMSG_DATA(cmhp)) = fd[i];
	  i++;
	  continue;
	} /* NRIGHTS*/

      /* Set message header to describe ancillary data NCREDS *IDs
	 that we want to send */
      if ( j < NCREDS)
	{
	  /* Construct an explicit credentials structure */
#ifdef __linux__
	  printf ("\nscm_rights+creds_send.c: Sending SCM_CREDENTIALS data\n");
	  cmhp->cmsg_len = clen2;
	  printf ("scm_rights+creds_send.c: cmhp[%d]->cmsg_len = %d\n", i+j, clen2);
	  sum_controllen += cspc2;
	  printf ("scm_rights+creds_send.c: CMSG_SPACE (sz)[%d] = %d\n", i+j, cspc2);

	  cmhp->cmsg_type = SCM_CREDENTIALS;
	  ucp = (struct ucred *) CMSG_DATA (cmhp);
	  /* Linux note: Sending wrong credentials results in EPERM */
	  ucp->pid = getpid();
	  ucp->uid = getuid();
	  ucp->gid = getgid();
	  printf("Sent credentials: pid=%ld, uid=%ld, euid=%d (from geteuid), gid=%ld\n",
		 (long) ucp->pid, (long) ucp->uid, geteuid(), (long) ucp->gid);
#else
	  printf ("\nscm_rights+creds_send.c: Sending SCM_CREDS data\n");
	  cmhp->cmsg_len = clen2;
	  printf ("scm_rights+creds_send.c: cmhp[%d]->cmsg_len = %d\n", i+j, clen2);
	  sum_controllen += cspc2;
	  printf ("scm_rights+creds_send.c: CMSG_SPACE (sz)[%d] = %d\n", i+j, cspc2);

	  cmhp->cmsg_type = SCM_CREDS;
	  ucp = (struct cmsgcred *) CMSG_DATA (cmhp);
	  /* GNU and kFreeBSD note: Sending any credentials results in
	     correct credentials received */
	  ucp->cmcred_pid = getpid();
	  ucp->cmcred_uid = getuid();
	  ucp->cmcred_euid = geteuid();
	  ucp->cmcred_gid = getgid();
	  printf("Sent credentials: cmcred_pid=%ld, cmcred_uid=%ld, cmcred_euid=%ld, cmcred_gid=%ld\n",
		 (long) ucp->cmcred_pid, (long) ucp->cmcred_uid, (long) ucp->cmcred_euid, (long) ucp->cmcred_gid);
#endif
	  j++;
	  continue;
	} /* NCREDS */
    } /* cmhp = ... */

  msgh.msg_controllen = sum_controllen;
  printf ("scm_rights+creds_send.c: msgh.msg_controllen = %d\n",(int)msgh.msg_controllen);

  addr.sun_family = AF_UNIX;
  strncpy(addr.sun_path, SOCK_PATH, sizeof(addr.sun_path) - 1);

  sfd = socket (AF_UNIX, SOCK_DGRAM, 0);
  if (sfd == -1)
    {
      perror("socket");
      exit(EXIT_FAILURE);
    }

if (connect(sfd, (struct sockaddr *) &addr,
	    sizeof(struct sockaddr_un)) == -1)
  {
    close(sfd);
    perror ("connect");
    exit(EXIT_FAILURE);
  }

 printf("Sending via datagram socket\n");
 ns = sendmsg(sfd, &msgh, 0);
 if (ns == -1)
   {
     perror("sendmsg");
     exit(EXIT_FAILURE);
   }
 printf("sendmsg() sent %ld bytes\n\n", (long) ns);
 unlink(SOCK_PATH);

 exit(EXIT_SUCCESS);
}
